/*
  Diamond Girl - Game where player collects diamonds.
  Copyright (C) 2005  Joni Yrjana
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


  Complete license can be found in the LICENSE file.
*/

#include "diamond_girl.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <SDL/SDL.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <string.h>
#include <errno.h>


int fullscreen;

#if !defined(AMIGA)
static char hdname[1024];
#endif


static int parse_arguments(int argc, char * argv[]);
static void settings_read(void);
static void settings_write(void);


const char * get_save_filename(const char * name)
{
  static char buf[1024 * 4];

#if defined(AMIGA)
  snprintf(buf, sizeof buf, "PROGDIR:%s", name);
#else
  snprintf(buf, sizeof buf, "%s/%s", hdname, name);
#endif
  
  return buf;
}

const char * get_data_filename(const char * name)
{
  static char buf[1024 * 4];

#if defined(AMIGA)
  snprintf(buf, sizeof buf, "PROGDIR:%s", name);
#else
  snprintf(buf, sizeof buf, "%s/%s", DIAMOND_GIRL_INSTALL_DIR, name);
#endif
  
  return buf;
}

static int parse_arguments(int argc, char * argv[])
{
  int rv, print_usage;

  rv = 1;
  print_usage = 0;

  for(int i = 1; i < argc; i++)
    {
      if(!strcmp(argv[i], "--no-sound") || !strcmp(argv[i], "-s"))
	use_sound = 0;
      else if(!strcmp(argv[i], "--sound") || !strcmp(argv[i], "-S"))
	use_sound = 1;
      else if(!strcmp(argv[i], "--fullscreen") || !strcmp(argv[i], "-f"))
	fullscreen = 1;
      else if(!strcmp(argv[i], "--windowed") || !strcmp(argv[i], "-w"))
	fullscreen = 0;
      else if(!strcmp(argv[i], "--help") || !strcmp(argv[i], "-h"))
	{
	  rv = 0;
	  print_usage = 1;
	}
      else
	{
	  fprintf(stderr, "%s: Invalid argument %d '%s'.\n", argv[0], (int) i, argv[i]);
	  rv = 0;
	  print_usage = 1;
	}
    }

  if(print_usage)
    {
      printf("Usage: %s [options]\n", argc > 0 ? argv[0] : "Diamond Girl");
      printf("  -h | --help        Print this help text and exit.\n");
      printf("  -s | --no-sound    Disable sounds and music.\n");
      printf("  -S | --sound       Enable sounds and music.\n");
      printf("  -f | --fullscreen  Fullscreen mode.\n");
      printf("  -w | --windowed    Windowed mode.\n");
    }

  return rv;
}


static void settings_read(void)
{ /* read settings */
  FILE * fp;
	
  fp = fopen(get_save_filename("settings"), "r");
  if(fp != NULL)
    {
      int done;
	    
      done = 0;
      while(!done)
	{
	  char linebuf[1024];
		
	  if(fgets(linebuf, sizeof linebuf, fp) != NULL)
	    {
	      char * lf;
	      char var[1024], val[1024];
	      int i, j;
		    
	      lf = strchr(linebuf, '\n');
	      if(lf != NULL)
		*lf = '\0';
		    
	      for(i = 0, j = 0; linebuf[i] != '\0' && linebuf[i] != '='; i++)
		var[j++] = linebuf[i];
	      var[j] = '\0';
	      if(linebuf[i++] != '\0')
		{
		  for(j = 0; linebuf[i] != '\0'; i++)
		    val[j++] = linebuf[i];
		  val[j] = '\0';
			
		  if(!strcmp(var, "FULLSCREEN"))
		    fullscreen = strtoul(val, NULL, 0);
		  else if(!strcmp(var, "SOUND"))
		    use_sound = strtoul(val, NULL, 0);
		  else if(!strcmp(var, "MAP_SELECTION"))
		    map_selection = strtol(val, NULL, 0);
		}
	    }
	  else
	    {
	      if(!feof(fp))
		fprintf(stderr, "Warning! An error occured while reading settings from '%s': %s\n", get_save_filename("settings"), strerror(errno));
	      done = 1;
	    }
	}
      fclose(fp);
    }
  else
    fprintf(stderr, "Warning! Failed to open settings file '%s': %s\n", get_save_filename("settings"), strerror(errno));
}
      

static void settings_write(void)
{ /* write settings */
  FILE * fp;

  fp = fopen(get_save_filename("settings"), "w");
  if(fp != NULL)
    {
      char linebuf[1024];
		    
      snprintf(linebuf, sizeof linebuf, "FULLSCREEN=%d\n", (int) fullscreen);
      if(fwrite(linebuf, strlen(linebuf), 1, fp) != 1)
	fprintf(stderr, "Failed to save settings into '%s': %s\n", get_save_filename("settings"), strerror(errno));

      snprintf(linebuf, sizeof linebuf, "SOUND=%d\n", (int) use_sound);
      if(fwrite(linebuf, strlen(linebuf), 1, fp) != 1)
	fprintf(stderr, "Failed to save settings into '%s': %s\n", get_save_filename("settings"), strerror(errno));
		    
      snprintf(linebuf, sizeof linebuf, "MAP_SELECTION=%d\n", (int) map_selection);
      if(fwrite(linebuf, strlen(linebuf), 1, fp) != 1)
	fprintf(stderr, "Failed to save settings into '%s': %s\n", get_save_filename("settings"), strerror(errno));
		    
      if(fclose(fp) != 0)
	fprintf(stderr, "Failed to save settings into '%s': %s\n", get_save_filename("settings"), strerror(errno));
    }
  else
    fprintf(stderr, "Failed to save settings into '%s': %s\n", get_save_filename("settings"), strerror(errno));
}


int main(int argc, char * argv[])
{
  int rv;
  int homedir_ok;

  use_sound = 1;
  fullscreen = 0;
  map_selection = 0;

#if defined(AMIGA)
  homedir_ok = 1;
#else
  { /* setup working directory to $HOME/.diamond_girl, making sure it exists, set homedir_ok variable accordingly */
    char * hd;

    homedir_ok = 0;
    hd = getenv("HOME");
    if(hd != NULL)
      {
	struct stat sb;

	snprintf(hdname, sizeof hdname, "%s/.diamond_girl", hd);
	mkdir(hdname, S_IREAD | S_IWRITE | S_IEXEC);
	if(stat(hdname, &sb) == 0)
	  homedir_ok = 1;
	else
	  fprintf(stderr, "Failed to setup working directory %s.\n", hdname);
      }
    else
      fprintf(stderr, "Failed to get environment variable HOME: %s\n", strerror(errno));
  }
#endif

  if(homedir_ok)
    {
      settings_read();
      if(parse_arguments(argc, argv))
	{
	  if(SDL_Init(0) == 0)
	    {
	      if(gfx_initialize())
		{
		  sfx_initialize();
		  main_menu();
		  rv = EXIT_SUCCESS;

		  settings_write();

		  sfx_cleanup();
		}
	      else
		rv = EXIT_FAILURE;

	      gfx_cleanup();
	    }
	  else
	    {
	      fprintf(stderr, "Failed to initialize SDL: %s\n", SDL_GetError());
	      rv = EXIT_FAILURE;
	    }

	  SDL_Quit();
	}
      else
	{
	  rv = EXIT_FAILURE;
	}
    }
  else
    {
      fprintf(stderr, "Failed to setup savedir.\n");
      rv = EXIT_FAILURE;
    }

  return rv;
}
