/*
  Diamond Girl - Game where player collects diamonds.
  Copyright (C) 2005  Joni Yrjana
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


  Complete license can be found in the LICENSE file.
*/

#include "diamond_girl.h"
#include "BFont.h"
#include "ui.h"

#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <SDL/SDL_framerate.h>
#include <SDL/SDL_gfxPrimitives.h>
#include <sys/types.h>
#include <dirent.h>
#include <ctype.h>

static int title_screen_selection;

static void draw_title_screen(void);
static int  title_screen_event(SDL_Event * event);
static void save_map_set_data(void);
static void load_map_set_data(void);

unsigned int get_rand(unsigned int max)
{
  return (float) max * (float) rand() / (float) (RAND_MAX + 1.0f);
}

static int quit;

static char * credits;
static SDL_Surface * credits_surface;
static size_t credits_pos;
static char * default_credits = "Diamond Girl - Created by Joni Yrjana                                                     ";
static int help_screen, help_screen_timer;
static int anim_frame, anim_timer;
static int highscores_first;

static int redraw_title;
static int redraw_help;
static int redraw_highscores;

static char ** map_names;
int            map_selection;
int            level_selection;
int            max_level_selection;
static int level_selection_change;

static struct ui_object * uiobj_newgame;
static struct ui_object * uiobj_mapset;
static struct ui_object * uiobj_mapset_left;
static struct ui_object * uiobj_mapset_right;
static struct ui_object * uiobj_level;
static struct ui_object * uiobj_level_left;
static struct ui_object * uiobj_level_right;
static struct ui_object * uiobj_editor;
static struct ui_object * uiobj_quit;

static void setup_ui(void);
static void button_newgame(struct ui_object * this, enum UI_BUTTON button);
static void button_newgame_draw(struct ui_object * this);
static void button_mapset_draw(struct ui_object * this);
static void button_mapset_left(struct ui_object * this, enum UI_BUTTON button);
static void button_mapset_left_draw(struct ui_object * this);
static void button_mapset_right(struct ui_object * this, enum UI_BUTTON button);
static void button_mapset_right_draw(struct ui_object * this);
static void button_level_draw(struct ui_object * this);
static void button_level_left(struct ui_object * this, enum UI_BUTTON button);
static void button_level_left_draw(struct ui_object * this);
static void button_level_right(struct ui_object * this, enum UI_BUTTON button);
static void button_level_right_draw(struct ui_object * this);
static void button_editor(struct ui_object * this, enum UI_BUTTON button);
static void button_editor_draw(struct ui_object * this);
static void button_quit(struct ui_object * this, enum UI_BUTTON button);
static void button_quit_draw(struct ui_object * this);

static int qsort_string_compare(const char ** a, const char ** b)
{
  return strcmp(*a, *b);
}


void main_menu(void)
{
  FPSmanager framerate_manager;
  int level_selection_change_timer;

  SDL_initFramerate(&framerate_manager);
  SDL_setFramerate(&framerate_manager, 50);

  quit = 0;
  title_screen_selection = 0;
  help_screen = 1;
  help_screen_timer = 0;
  anim_frame = 0;
  anim_timer = 0;
  highscores_first = 0;

  redraw_title = 1;
  redraw_help = help_screen + 1;
  redraw_highscores = 1;

  map_names = NULL;

  level_selection_change = 0;

  { /* read credits */
    FILE * fp;

    credits = NULL;
    fp = fopen(get_data_filename("CREDITS"), "r");
    if(fp != NULL)
      {
	int f;
	struct stat sb;

	f = fileno(fp);
	if(fstat(f, &sb) == 0)
	  {
	    /* calculate linefeeds */
	    int c;
	    int lf;
	    const char * lf_text = "      ";

	    lf = 0;
	    do
	      {
		c = fgetc(fp);
		if(c == '\n')
		  lf++;
	      } while(c != EOF);
		  
	    rewind(fp);

	    credits = malloc(strlen(lf_text) * 10 + sb.st_size + lf * strlen(lf_text) + 1);
	    assert(credits != NULL);

	    { /* read the credits file and replace linefeeds with lf_text */
	      char * p;

	      p = credits;
	      for(int i = 0; i < 10; i++)
		{
		  memcpy(p, lf_text, strlen(lf_text));
		  p += strlen(lf_text);
		}

	      do
		{
		  c = fgetc(fp);
		  if(c != EOF)
		    {
		      if(c == '\n')
			{
			  memcpy(p, lf_text, strlen(lf_text));
			  p += strlen(lf_text);
			}
		      else
			{
			  *p = c;
			  p++;
			}
		    }
		} while(c != EOF);
	      *p = '\0';
	    }
	  }
	else
	  {
	    fprintf(stderr, "Failed to stat 'CREDITS': %s\n", strerror(errno)); 
	  }
	fclose(fp);
      }
    else
      {
	fprintf(stderr, "Failed to open 'CREDITS': %s\n", strerror(errno));
      }
	
    if(credits == NULL)
      credits = strdup(default_credits);
  }

  assert(credits != NULL);
  credits_surface = NULL;

  { /* read map names */
    DIR * d;

    d = opendir(get_data_filename("maps"));
    if(d != NULL)
      {
	struct dirent * e;
	size_t s;

	s = 0;
	do
	  {
	    e = readdir(d);
	    if(e != NULL)
	      {
		if(isalnum(e->d_name[0]))
		  {
		    char ** tmp;

		    tmp = realloc(map_names, sizeof(char *) * (s + 1 + 1));
		    if(tmp != NULL)
		      {
			map_names = tmp;
			map_names[s] = strdup(e->d_name);
			if(map_names[s] == NULL)
			  {
			    fprintf(stderr, "Failed to allocate memory for predefined map names: %s\n", strerror(errno));
			    e = NULL;
			  }
			s++;
			map_names[s] = NULL;
		      }
		    else
		      {
			fprintf(stderr, "Failed to allocate memory for predefined map names: %s\n", strerror(errno));
			e = NULL;
		      }
		  }
	      }
	  }
	while(e != NULL);
	closedir(d);
      }
    else
      fprintf(stderr, "Failed to open directory '%s': %s\n", get_data_filename("maps"), strerror(errno));
  }
  
  if(map_names != NULL)
    {
      int n;
      
      n = 0;
      while(map_names[n] != NULL)
	n++;

      if(map_selection > n)
	map_selection = n - 1;

      qsort(map_names, n, sizeof(char *), (int(*)(const void *, const void *)) qsort_string_compare);
    }
  else
    {
      fprintf(stderr, "Warning, no predefined maps available.\n");
      map_selection = -1;
    }

  load_map_set_data();

  clear_area(0, 0, gfx(MAP_SCREEN)->w, gfx(MAP_SCREEN)->h);

  setup_ui();
  sfx_music(MUSIC_TITLE, 1);

  level_selection_change_timer = 0;

  gfx_frame0();

  while(!quit)
    {
      SDL_Event event;
      int process_more;

      process_more = 1;
      while(SDL_PollEvent(&event) && process_more)
	process_more = title_screen_event(&event);

      level_selection_change_timer++;
      if(level_selection_change_timer >= 5)
	{
	  level_selection_change_timer = 0;
	  if(level_selection_change && level_selection + level_selection_change >= 1 && level_selection + level_selection_change <= max_level_selection)
	    {
	      level_selection += level_selection_change;
	      ui_object_set_dirty(uiobj_level);
	    }
	}

      if(!quit)
	{
	  draw_title_screen();
	  SDL_Flip(gfx(MAP_SCREEN));
	  SDL_framerateDelay(&framerate_manager);
	}
    }

  save_map_set_data();

  ui_cleanup();
}

void clear_area(int x, int y, int w, int h)
{
  SDL_Rect r;
  Uint32 c;

  c = SDL_MapRGB(gfx(MAP_SCREEN)->format, 0, 0, 0);
  r.x = x;
  r.y = y;
  r.w = w;
  r.h = h;
  SDL_FillRect(gfx(MAP_SCREEN), &r, c);
}

static void draw_title_screen(void)
{
  int font_height, y;

  anim_timer++;
  if(anim_timer >= 10)
    {
      anim_frame++;
      if(anim_frame >= 0xffff)
	anim_frame = 0;
      anim_timer = 0;
      gfx_next_frame();
      redraw_help = help_screen - 1;
    }

  font_height = BFont_FontHeight(BFont_GetCurrentFont());
  y = 0;

  if(redraw_title)
    { /* title */
      clear_area(0, y, gfx(MAP_SCREEN)->w, font_height);
      BFont_CenteredPrintString(gfx(MAP_SCREEN), y, "DIAMOND GIRL v%d.%d", (int) DIAMOND_GIRL_VERSION, (int) DIAMOND_GIRL_REVISION);
      redraw_title = 0;
    }
  y += font_height * 2;

  if(redraw_help != help_screen)
    { /* help/instructions */
      redraw_help = help_screen;

      clear_area(gfx(MAP_SCREEN)->w / 2, y, gfx(MAP_SCREEN)->w / 2, gfx(MAP_SCREEN)->h - font_height * 2 - (font_height + 5) * 3 - y);
      switch(help_screen)
	{
	case 0: /* description */
	case 2: /* hints */
	  {
	    const char * description[] =
	      {
		"Welcome to Diamond Girl.",
		"",
		"Collect diamonds to open exit",
		"  to the next level.",
		"",
		"Touching any of the 'animals' will kill you.",
		"Falling boulders and diamonds",
		"  will also kill you.",
		"",
		"Controls:",
		" * cursor keys to move",
		" * shift+cursor keys to manipulate grid",
		"   without moving",
		" * escape key to quit",
		" * F1 for ingame help and pause",
		" * F10 toggle windowed/fullscreen",
		NULL
	      };
	    const char * hints[] =
	      {
		"Hints:",
		"",
		"You can push boulders left and right",
		"  if there is empty space behind them.",
		"",
		"Pushing speeds up a bit after first step.",
		"",
		"Drop boulders or diamonds on top of",
		"  animals to kill them.",
		"Watch out for the explosion!",
		"",
		"Each butterfly explodes into nine diamonds,",
		"  and bugs don't leave anything behind.",
		"",
		"You will get extra girl every 200 score.",
		NULL
	      };
	    const char ** text;
	    int i, x;

	    i = 0;
	    x = gfx(MAP_SCREEN)->w / 2;
	    if(help_screen == 0)
	      text = description;
	    else // if(help_screen == 2)
	      text = hints;
	    while(text[i] != NULL)
	      {
		BFont_PutString(gfx(MAP_SCREEN), x, y, text[i]);
		y += font_height;
		i++;
	      }
	  }
	  break;
	case 1: /* game graphics and explanations of them */
	  {
	    struct
	    {
	      enum MAP_GLYPH glyph;
	      const char *   explanation;
	    } explanations[] = 
	      {
		{ MAP_PLAYER_ARMOUR3,    "You." },
		{ MAP_DIAMOND,           "A diamond, collect." },
		{ MAP_BOULDER,           "A boulder." },
		{ MAP_SAND,              "Sand." },
		{ MAP_EXTRA_LIFE_ANIM,   "For few seconds, empty space looks like" },
		{ MAP_EXTRA_LIFE_ANIM,   "  this when you get extra girl." },
		{ MAP_BRICK,             "A brick wall." },
		{ MAP_BRICK_MORPHER,     "A morphing wall, turns diamonds into" },
		{ MAP_BRICK_MORPHER,     "  boulders and boulders into diamonds." },
		{ MAP_BRICK_UNBREAKABLE, "Unbreakable brick wall." },
		{ MAP_AMEBA,             "Ameba." },
		{ MAP_EXIT_LEVEL,        "Exit to the next level." },
		{ MAP_ENEMY1,            "A radioactive bug, avoid." },
		{ MAP_ENEMY2,            "A radioactive butterfly, avoid." },
		{ MAP_SCREEN,            NULL }
	      };
	    int i;
	    
	    i = 0;
	    while(explanations[i].explanation != NULL)
	      {
		gfx_draw_glyph(gfx(MAP_SCREEN)->w / 2, y + font_height, explanations[i].glyph, MOVE_NONE);
		
		if(font_height > 26)
		  y += font_height;
		else
		  y += 26;
		
		BFont_PutString(gfx(MAP_SCREEN), gfx(MAP_SCREEN)->w / 2 + 30, y, explanations[i].explanation);
		
		i++;
	      }
	  }
	  break;
	}
    }

  help_screen_timer++;
  if(help_screen_timer >= 50 * 12)
    {
      help_screen++;
      if(help_screen >= 3)
	help_screen = 0;
      help_screen_timer = 0;
    }

  ui_draw();

  /* credits */
  if(credits_surface == NULL)
    {
      SDL_Surface * tmp;
      
      tmp = BFont_CreateSurface(credits);
      if(tmp != NULL)
	{
	  credits_surface = SDL_ConvertSurface(tmp, gfx(MAP_SCREEN)->format, SDL_HWSURFACE);
	  if(credits_surface != NULL)
	    SDL_FreeSurface(tmp);
	  else
	    credits_surface = tmp;
	}
    }


  if(credits_surface != NULL)
    {
      SDL_Rect sr, dr;

      dr.x = gfx(MAP_SCREEN)->w / 8;
      dr.y = gfx(MAP_SCREEN)->h - font_height;
      dr.w = gfx(MAP_SCREEN)->w - dr.x * 2;
      dr.h = credits_surface->h;

      sr.x = credits_pos;
      sr.y = 0;
      sr.w = dr.w;
      sr.h = credits_surface->h;

      clear_area(dr.x, dr.y, dr.w, dr.h);
      
      SDL_BlitSurface(credits_surface, &sr, gfx(MAP_SCREEN), &dr);

      credits_pos++;
      if(credits_pos >= (size_t) credits_surface->w)
	credits_pos = 0;
    }


  if(redraw_highscores)
    { /* highscore list */
      size_t size, amount;
      struct highscore_entry ** entries;
     
      redraw_highscores = 0;
      y = font_height * 2;
      amount = 20;
      clear_area(0, y, gfx(MAP_SCREEN)->w / 2, (1 + amount + 1) * font_height);
      BFont_PrintString(gfx(MAP_SCREEN), 10, y, "Highscores %d -> %d:  (PageUp/Down)", highscores_first + 1, highscores_first + amount);
      y += font_height;
      hlineRGBA(gfx(MAP_SCREEN), 0, gfx(MAP_SCREEN)->w / 2 - 20, y, 0x00, 0xff, 0x00, 0xff);
      vlineRGBA(gfx(MAP_SCREEN), 40, y, y + (amount + 1) * font_height, 0x00, 0xff, 0x00, 0xff);
      vlineRGBA(gfx(MAP_SCREEN), 130, y, y + (amount + 1) * font_height, 0x00, 0xff, 0x00, 0xff);
      vlineRGBA(gfx(MAP_SCREEN), 190, y, y + (amount + 1) * font_height, 0x00, 0xff, 0x00, 0xff);
      BFont_PrintString(gfx(MAP_SCREEN), 10, y, " #");
      BFont_PrintString(gfx(MAP_SCREEN), 40, y, " Score");
      BFont_PrintString(gfx(MAP_SCREEN), 130, y, " Level");
      BFont_PrintString(gfx(MAP_SCREEN), 190, y, " Timestamp");
      y += font_height;
      hlineRGBA(gfx(MAP_SCREEN), 0, gfx(MAP_SCREEN)->w / 2 - 20, y, 0x00, 0xff, 0x00, 0xff);
      entries = highscores_get(&size);
      for(size_t i = highscores_first; i < highscores_first + amount && i < size; i++)
	{
	  char buf[256];
	  struct tm * tm;
	  
	  BFont_PrintString(gfx(MAP_SCREEN), 10, y, "%3d", (int) i + 1);
	  BFont_PrintString(gfx(MAP_SCREEN), 40, y, " %6d", (int) entries[i]->score);
	  BFont_PrintString(gfx(MAP_SCREEN), 130, y, " %3d", (int) entries[i]->level);
	  
	  tm = localtime(&entries[i]->timestamp);
	  strftime(buf, sizeof buf, "%Y-%m-%d %H:%M:%S", tm);
	  BFont_PrintString(gfx(MAP_SCREEN), 190, y, " %s", buf);
	  
	  y += font_height;
	}
    }
}





static int title_screen_event(SDL_Event * event)
{
  int process_more;

  process_more = 1;
  switch(event->type)
    {
    case SDL_QUIT:
      quit = 1;
      break;
    case SDL_KEYDOWN:
      if(event->key.keysym.sym == SDLK_ESCAPE)
	{
	  quit = 1;
	}
      else if(event->key.keysym.sym == SDLK_RETURN || event->key.keysym.sym == SDLK_KP_ENTER)
	{
	  struct ui_object * obj;

	  obj = ui_object_selected();
	  
	  if(obj == NULL)
	    obj = uiobj_newgame;

	  if(obj != NULL)
	    {
	      ui_func_activate_t func;
	      
	      func = ui_object_activate_function(obj);
	      if(func != NULL)
		func(obj, UI_BUTTON_LEFT);
	    }
	}
      else if(event->key.keysym.sym == SDLK_DOWN)
	{
	  struct ui_object * obj;

	  obj = ui_object_selected();
	  if(obj == NULL)
	    obj = uiobj_newgame;

	  if(obj == uiobj_newgame)
	    obj = uiobj_editor;
	  else if(obj == uiobj_editor)
	    obj = uiobj_quit;

	  if(obj != NULL)
	    {
	      ui_object_set_selected(obj);
	      ui_draw();
	    }
	}
      else if(event->key.keysym.sym == SDLK_UP)
	{
	  struct ui_object * obj;

	  obj = ui_object_selected();
	  if(obj == NULL)
	    obj = uiobj_newgame;

	  if(obj == uiobj_editor)
	    obj = uiobj_newgame;
	  else if(obj == uiobj_quit)
	    obj = uiobj_editor;

	  if(obj != NULL)
	    {
	      ui_object_set_selected(obj);
	      ui_draw();
	    }
	}
      else if(event->key.keysym.sym == SDLK_LEFT)
	{
	  if(map_selection > -1)
	    {
	      save_map_set_data();

	      map_selection--;
	      ui_object_set_dirty(uiobj_mapset);
	      ui_object_set_dirty(uiobj_level);
	      ui_object_set_dirty(uiobj_editor);
	      redraw_highscores = 1;
	      highscores_first = 0;

	      load_map_set_data();
	    }
	}
      else if(event->key.keysym.sym == SDLK_RIGHT)
	{
	  if(map_names != NULL && map_names[map_selection + 1] != NULL)
	    {
	      save_map_set_data();
	      
	      map_selection++;
	      ui_object_set_dirty(uiobj_mapset);
	      ui_object_set_dirty(uiobj_level);
	      ui_object_set_dirty(uiobj_editor);
	      redraw_highscores = 1;
	      highscores_first = 0;

	      load_map_set_data();
	    }
	}
      else if(event->key.keysym.sym == SDLK_PAGEUP)
	{
	  if(highscores_first > 0)
	    highscores_first -= 20;
	  if(highscores_first < 0)
	    highscores_first = 0;
	  redraw_highscores = 1;
	}
      else if(event->key.keysym.sym == SDLK_PAGEDOWN)
	{
	  size_t size;

	  highscores_get(&size);
	  if(highscores_first + 20 < (int) size)
	    highscores_first += 20;
	  redraw_highscores = 1;
	}
      else if(event->key.keysym.sym == SDLK_F10)
	{
	  if(fullscreen)
	    fullscreen = 0;
	  else
	    fullscreen = 1;
	  gfx_cleanup();
	  gfx_initialize();
	  redraw_title = 1;
	  redraw_help = help_screen + 1;
	  ui_object_set_dirty(NULL);
	  redraw_highscores = 1;
	}
      else if(event->key.keysym.sym == SDLK_COMMA)
	{
	  level_selection_change = -1;
	  process_more = 0;
	}
      else if(event->key.keysym.sym == SDLK_PERIOD)
	{
	  level_selection_change = 1;
	  process_more = 0;
	}
      break;
    case SDL_KEYUP:
      if(event->key.keysym.sym == SDLK_COMMA && level_selection_change == -1)
	level_selection_change = 0;
      else if(event->key.keysym.sym == SDLK_PERIOD && level_selection_change == 1)
	level_selection_change = 0;
      break;
    case SDL_MOUSEBUTTONDOWN:
      {
	enum UI_BUTTON b;
	
	if(event->button.button == SDL_BUTTON_LEFT)
	  b = UI_BUTTON_LEFT;
	else if(event->button.button == SDL_BUTTON_MIDDLE)
	  b = UI_BUTTON_MIDDLE;
	else if(event->button.button == SDL_BUTTON_RIGHT)
	  b = UI_BUTTON_RIGHT;
	else
	  b = UI_BUTTON_LEFT;
	
	ui_process_mouse_button(event->button.x, event->button.y, b, 0);
      }
      break;
    case SDL_MOUSEBUTTONUP:
      {
	enum UI_BUTTON b;
	
	if(event->button.button == SDL_BUTTON_LEFT)
	  b = UI_BUTTON_LEFT;
	else if(event->button.button == SDL_BUTTON_MIDDLE)
	  b = UI_BUTTON_MIDDLE;
	else if(event->button.button == SDL_BUTTON_RIGHT)
	  b = UI_BUTTON_RIGHT;
	else
	  b = UI_BUTTON_LEFT;
	
	ui_process_mouse_button(event->button.x, event->button.y, b, 1);
      }
      break;
    case SDL_MOUSEMOTION:
      ui_process_mouse_move(event->motion.x, event->motion.y);
      break;
    }

  ui_draw();

  return process_more;
}



static void load_map_set_data(void)
{
  char * map_set;

  if(map_selection >= 0)
    map_set = map_names[map_selection];
  else
    map_set = NULL;

  highscores_load(map_set);
  
  {
    FILE * fp;
    char fn[1024];
    
    snprintf(fn, sizeof fn, "data-%s", map_set != NULL ? map_set : "");
    fp = fopen(get_save_filename(fn), "rb");
    if(fp != NULL)
      {
	if(fread(&max_level_selection, sizeof max_level_selection, 1, fp) != 1)
	  {
	    fprintf(stderr, "Warning! Failed to read from '%s': %s\n", get_save_filename(fn), strerror(errno));
	    max_level_selection = 1;
	  }
	fclose(fp);
      }
    else
      {
	fprintf(stderr, "Warning! Failed to open '%s' for reading: %s\n", get_save_filename(fn), strerror(errno));
	max_level_selection = 1;
      }
  }

  level_selection = 1;
}


static void save_map_set_data(void)
{
  char * map_set;

  if(map_selection >= 0)
    map_set = map_names[map_selection];
  else
    map_set = NULL;

  {
    FILE * fp;
    char fn[1024];

    snprintf(fn, sizeof fn, "data-%s", map_set != NULL ? map_set : "");
    fp = fopen(get_save_filename(fn), "wb");
    if(fp != NULL)
      {
	if(fwrite(&max_level_selection, sizeof max_level_selection, 1, fp) != 1)
	  fprintf(stderr, "Warning! Failed to write to '%s': %s\n", get_save_filename(fn), strerror(errno));
	if(fclose(fp) != 0)
	  fprintf(stderr, "Warning! Failed to write to '%s': %s\n", get_save_filename(fn), strerror(errno));
      }
    else
      fprintf(stderr, "Warning! Failed to open '%s' for writing: %s\n", get_save_filename(fn), strerror(errno));
  }
}

static void setup_ui(void)
{
  int font_height;
  int x, y, w, h;

  ui_initialize();
  
  font_height = BFont_FontHeight(BFont_GetCurrentFont());
  x = 10;
  y = gfx(MAP_SCREEN)->h - font_height * 2 - (font_height + 5) * 3;
  w = 200;
  h = font_height + 4;

  uiobj_newgame = ui_object(x, y, w, h);
  assert(uiobj_newgame != NULL);
  ui_object_set_draw_function(uiobj_newgame, button_newgame_draw);
  ui_object_set_activate_function(uiobj_newgame, button_newgame);
  ui_object_set_selected(uiobj_newgame);

  uiobj_mapset_left = ui_object(x + w + 10, y, 30, h);
  assert(uiobj_mapset_left != NULL);
  ui_object_set_draw_function(uiobj_mapset_left, button_mapset_left_draw);
  ui_object_set_activate_function(uiobj_mapset_left, button_mapset_left);

  uiobj_mapset_right = ui_object(x + w + 10 + 30, y, 30, h);
  assert(uiobj_mapset_right != NULL);
  ui_object_set_draw_function(uiobj_mapset_right, button_mapset_right_draw);
  ui_object_set_activate_function(uiobj_mapset_right, button_mapset_right);
  
  uiobj_mapset = ui_object(x + w + 10 + 30 + 30, y, 400, h);
  assert(uiobj_mapset != NULL);
  ui_object_set_draw_function(uiobj_mapset, button_mapset_draw);

  y += font_height + 5;

  uiobj_editor = ui_object(x, y, w, h);
  assert(uiobj_editor != NULL);
  ui_object_set_draw_function(uiobj_editor, button_editor_draw);
  ui_object_set_activate_function(uiobj_editor, button_editor);

  uiobj_level_left = ui_object(x + w + 10, y, 30, h);
  assert(uiobj_level_left != NULL);
  ui_object_set_draw_function(uiobj_level_left, button_level_left_draw);
  ui_object_set_activate_function(uiobj_level_left, button_level_left);

  uiobj_level_right = ui_object(x + w + 10 + 30, y, 30, h);
  assert(uiobj_level_right != NULL);
  ui_object_set_draw_function(uiobj_level_right, button_level_right_draw);
  ui_object_set_activate_function(uiobj_level_right, button_level_right);
  
  uiobj_level = ui_object(x + w + 10 + 30 + 30, y, 400, h);
  assert(uiobj_level != NULL);
  ui_object_set_draw_function(uiobj_level, button_level_draw);

  y += font_height + 5;

  uiobj_quit = ui_object(x, y, w, h);
  assert(uiobj_quit != NULL);
  ui_object_set_draw_function(uiobj_quit, button_quit_draw);
  ui_object_set_activate_function(uiobj_quit, button_quit);
}


static void button_newgame_draw(struct ui_object * this)
{
  char * m;

  if(map_selection < 0)
    m = "random maps";
  else
    m = map_names[map_selection];
  button_draw_text(this, "New game");
}

static void button_newgame(struct ui_object * this, enum UI_BUTTON button)
{
  assert(this == this);

  if(button == UI_BUTTON_LEFT)
    {
      int max_level;
      char * map_set;
	  
      if(map_selection == -1)
	map_set = NULL;
      else
	map_set = map_names[map_selection];
      
      ui_cleanup();
      clear_area(0, 0, gfx(MAP_SCREEN)->w, gfx(MAP_SCREEN)->h);
      quit = game(map_set, level_selection, 0, &max_level);
      if(max_level > max_level_selection)
	max_level_selection = max_level;
      sfx_music(MUSIC_TITLE, 1);
      clear_area(0, 0, gfx(MAP_SCREEN)->w, gfx(MAP_SCREEN)->h);
      setup_ui();
      redraw_title = 1;
      redraw_help = help_screen + 1;
      redraw_highscores = 1;
      gfx_frame0();
    }
}

static void button_editor_draw(struct ui_object * this)
{
  if(map_selection < 0)
    object_draw_text(this, "Editor N/A");
  else
    button_draw_text(this, "Editor");
}

static void button_editor(struct ui_object * this, enum UI_BUTTON button)
{
  assert(this == this);
  assert(button == button);

  if(button == UI_BUTTON_LEFT)
    {
      char * map_set;
	  
      if(map_selection == -1)
	map_set = NULL;
      else
	map_set = map_names[map_selection];
      if(map_set != NULL)
	{
	  ui_cleanup();
	  clear_area(0, 0, gfx(MAP_SCREEN)->w, gfx(MAP_SCREEN)->h);
	  sfx_music_stop();
	  map_editor(map_set);
	  sfx_music(MUSIC_TITLE, 1);
	  clear_area(0, 0, gfx(MAP_SCREEN)->w, gfx(MAP_SCREEN)->h);
	  setup_ui();
	  redraw_title = 1;
	  redraw_help = help_screen + 1;
	  redraw_highscores = 1;
	  gfx_frame0();
	}
    }
}


static void button_quit_draw(struct ui_object * this)
{
  button_draw_text(this, "Quit");
}

static void button_quit(struct ui_object * this, enum UI_BUTTON button)
{
  assert(this == this);
  if(button == UI_BUTTON_LEFT)
    {
      quit = 1;
    }
}

static void button_mapset_draw(struct ui_object * this)
{
  char * map_set;
	  
  if(map_selection == -1)
    map_set = "random maps";
  else
    map_set = map_names[map_selection];

  object_draw_text(this, "Map set: %s", map_set);
}

static void button_mapset_left(struct ui_object * this, enum UI_BUTTON button)
{
  assert(this == this);
  assert(button == button);
  
  if(map_selection > -1)
    {
      save_map_set_data();

      map_selection--;
      ui_object_set_dirty(uiobj_mapset);
      ui_object_set_dirty(uiobj_level);
      ui_object_set_dirty(uiobj_editor);
      redraw_highscores = 1;
      highscores_first = 0;
      
      load_map_set_data();
    }
}

static void button_mapset_left_draw(struct ui_object * this)
{
  button_draw_text(this, " < ");
}

static void button_mapset_right(struct ui_object * this, enum UI_BUTTON button)
{
  assert(this == this);
  assert(button == button);

  if(map_names != NULL && map_names[map_selection + 1] != NULL)
    {
      save_map_set_data();

      map_selection++;
      ui_object_set_dirty(uiobj_mapset);
      ui_object_set_dirty(uiobj_level);
      ui_object_set_dirty(uiobj_editor);
      redraw_highscores = 1;
      highscores_first = 0;

      load_map_set_data();
    }
}

static void button_mapset_right_draw(struct ui_object * this)
{
  button_draw_text(this, " > ");
}

static void button_level_draw(struct ui_object * this)
{
  object_draw_text(this, "Level: %d / %d", (int) level_selection, (int) max_level_selection);
}

static void button_level_left(struct ui_object * this, enum UI_BUTTON button)
{
  assert(this == this);
  assert(button == button);
  
  if(level_selection > 1)
    {
      level_selection--;
      ui_object_set_dirty(uiobj_level);
    }
}

static void button_level_left_draw(struct ui_object * this)
{
  button_draw_text(this, " < ");
}

static void button_level_right(struct ui_object * this, enum UI_BUTTON button)
{
  assert(this == this);
  assert(button == button);

  if(level_selection < max_level_selection)
    {
      level_selection++;
      ui_object_set_dirty(uiobj_level);
    }
}

static void button_level_right_draw(struct ui_object * this)
{
  button_draw_text(this, " > ");
}
